#!/usr/bin/env python
# Copyright 2014 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.


# standard library modules, , ,
import unittest

# sourceparse, , parse version source urls, internal
from yotta.lib import sourceparse

# Shorthand URLs for GitHub
ShortHand_URLs = [
    'username/reponame',
]

# Longhand URLs for GitHub
Github_URLs = [
    'https://github.com/username/reponame.git',
    'git://github.com/username/reponame.git',
    'git+http://username@github.com/username/reponame.git',
    'git+https://username@github.com/username/reponame.git',
]

Git_URLs = [
    'http://somewhere.something/reponame.git',
    'https://somewhere.something/reponame.git',
    'ssh://somewhere.com/something/etc/reponame.git',
    'git+ssh://somewhere.com/something/etc/reponame',
    'git@github.com:username/reponame.git',
]

HG_URLs = [
    'http://somewhere.something/reponame.hg',
    'https://somewhere.something/reponame.hg',
    'ssh://somewhere.com/something/etc/reponame.hg',
    'hg+ssh://somewhere.com/something/etc/reponame',
]

# We support version spec, branch name, tag name and commit id for GitHub and Git
Git_Specs = [
    '',
    '1.2.3',
    '^1.2.3',
    '~1.2.3',
    '-1.2.3',
    'branch-or-tag-name',
    'branch+or+tag+name',
    'd5f5049',
]

# We support only version spec for HG
HG_Specs = [
    '',
    '1.2.3',
    '^1.2.3',
    '~1.2.3',
]

Registry_Specs = [
    '',
    '*',
    '1.2.3',
    '>=1.2.3',
    '^0.1.2',
    '~0.1.2',
]

test_invalid_urls = [
    'stuff otherstuff',
    'somthing/somethingelse/somethingelseagain'
]


class TestParseSourceURL(unittest.TestCase):
    def test_registryURLs(self):
        for url in Registry_Specs:
            sv = sourceparse.parseSourceURL(url)
            self.assertEqual(sv.source_type, 'registry')

    def test_shorthandURLs(self):
        for url in ShortHand_URLs:
            for s in Git_Specs:
                if len(s):
                    # Shorthand URLs support '@' and ' ' as well as '#'
                    for m in ['#', '@', ' ']:
                        sv = sourceparse.parseSourceURL(url + m + s)
                        self.assertEqual(sv.source_type, 'github')
                        self.assertEqual(sv.spec, s)
                else:
                    sv = sourceparse.parseSourceURL(url)
                    self.assertEqual(sv.source_type, 'github')
                    self.assertEqual(sv.spec, s)

    def test_githubURLs(self):
        for url in Github_URLs:
            for s in Git_Specs:
                if len(s):
                    source = url + '#' + s
                else:
                    source = url
                sv = sourceparse.parseSourceURL(source)
                self.assertEqual(sv.source_type, 'github')
                self.assertEqual(sv.spec, s)

    def test_gitURLs(self):
        for url in Git_URLs:
            for s in Git_Specs:
                if len(s):
                    source = url + '#' + s
                else:
                    source = url
                sv = sourceparse.parseSourceURL(source)
                self.assertEqual(sv.source_type, 'git')
                self.assertEqual(sv.spec, s)

    def test_hgURLs(self):
        for url in HG_URLs:
            for s in HG_Specs:
                if len(s):
                    source = url + '#' + s
                else:
                    source = url
                sv = sourceparse.parseSourceURL(source)
                self.assertEqual(sv.source_type, 'hg')
                self.assertEqual(sv.spec, s)

    def test_invalid(self):
        for url in test_invalid_urls:
            self.assertRaises(ValueError, sourceparse.parseSourceURL, url)

Valid_Names = [
    'validname',
    'another-validname',
    'another-valid2name123'
]

class TestParseModuleNameAndSpec(unittest.TestCase):
    def test_validNames(self):
        for name in Valid_Names:
            n, s = sourceparse.parseModuleNameAndSpec(name)
            self.assertEqual(n, name)
            self.assertEqual(s, '*')

    def test_ShorthandRefs(self):
        for url in ShortHand_URLs:
            for spec in Git_Specs:
                if len(spec):
                    # Shorthand URLs support '@' and ' ' as well as '#'
                    for m in ['#', '@', ' ']:
                        ns = url + m + spec
                        n, s = sourceparse.parseModuleNameAndSpec(ns)
                        self.assertEqual(n, 'reponame')
                        self.assertEqual(s, ns)
                else:
                    n, s = sourceparse.parseModuleNameAndSpec(url)
                    self.assertEqual(n, 'reponame')
                    self.assertEqual(s, url)

    def test_GithubRefs(self):
        for url in Github_URLs:
            for spec in Git_Specs:
                if len(spec):
                    ns = url + '#' + spec
                else:
                    ns = url
                n, s = sourceparse.parseModuleNameAndSpec(ns)
                self.assertEqual(n, 'reponame')
                self.assertEqual(s, ns)

    def test_GitRefs(self):
        for url in Git_URLs:
            for spec in Git_Specs:
                if len(spec):
                    ns = url + '#' + spec
                else:
                    ns = url
                n, s = sourceparse.parseModuleNameAndSpec(ns)
                self.assertEqual(n, 'reponame')
                self.assertEqual(s, ns)

    def test_HGRefs(self):
        for url in HG_URLs:
            for spec in HG_Specs:
                if len(spec):
                    ns = url + '#' + spec
                else:
                    ns = url
                n, s = sourceparse.parseModuleNameAndSpec(ns)
                self.assertEqual(n, 'reponame')
                self.assertEqual(s, ns)

    def test_atVersion(self):
        for name in Valid_Names:
            for v in Registry_Specs:
                if len(v):
                    nv = name + '@' + v
                    n, s = sourceparse.parseModuleNameAndSpec(nv)
                    self.assertEqual(n, name)
                    self.assertEqual(s, v)


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

# standard library options
from argparse import Action

# logging setup, , setup the logging system, internal
from yotta.lib import logging_setup

class PlainAction(Action):
    def __init__(self, *args, **kwargs):
        kwargs['nargs'] = 0
        kwargs['metavar'] = None
        self.dest = kwargs['dest']
        super(PlainAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        logging_setup.setPlain(True)

class ColourfulAction(PlainAction):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, False)
        logging_setup.setPlain(False)


def addTo(parser):
    parser.add_argument('--plain', dest='plain', action=PlainAction,
        default=logging_setup.plainOutputByDefault(), nargs=0, help="Use a simple output format with no colours."
    )
    parser.add_argument('--colourful', dest='plain', action=ColourfulAction,
        default=(not logging_setup.plainOutputByDefault()), nargs=0, help="Force colourful output, even if the output is not to a tty."
    )




# Copyright 2014 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

# Utilities for parsing version source URLs, and determining how the
# specified version should be fetched.

# standard library modules, , ,
try:
    from urlparse import urlsplit
except ImportError:
    from urllib.parse import urlsplit #pylint: disable=no-name-in-module,import-error

# version, , represent versions and specifications, internal
from yotta.lib import version


class InvalidVersionSpec(ValueError):
    pass

class VersionSource(object):
    def __init__(self, source_type, location, spec):
        assert(source_type in ('registry', 'github', 'git', 'hg'))
        self.source_type = source_type
        self.location = location
        self.spec = spec
        try:
            self.semantic_spec = version.Spec(spec)
        except ValueError:
            # for git/github source URLs the spec is allowed to be a branch
            # name or tag name, as well as a valid semantic version
            # specification
            # !!! TODO: also allow hg here
            if source_type in ('git', 'github'):
                self.semantic_spec = None
            else:
                raise ValueError(
                    "Invalid semantic version spec: \"%s\"" % spec
                )

    def semanticSpec(self):
        return self.semantic_spec or version.Spec('*')

    def semanticSpecMatches(self, v):
        if isinstance(v, str):
            v = version.Version(v)
        if self.semantic_spec is None:
            return True
        else:
            return self.semantic_spec.match(v)


def _getNonRegistryRef(source_url):
    import re

    # something/something#spec = github
    # something/something@spec = github
    # something/something spec = github
    github_match = re.match(r'^([.a-z0-9_-]+/([.a-z0-9_-]+)) *[@#]?([^/:\?\[\\]*)$', source_url, re.I)
    if github_match:
        return github_match.group(2), VersionSource('github', github_match.group(1), github_match.group(3))

    parsed = urlsplit(source_url)

    # github
    if parsed.netloc.endswith('github.com'):
        # any URL onto github should be fetched over the github API, even if it
        # would parse as a valid git URL
        name_match = re.match('^/([.a-z0-9_-]+/([.a-z0-9_-]+?))(.git)?$', parsed.path, re.I)
        if name_match:
            return name_match.group(2), VersionSource('github', name_match.group(1), parsed.fragment)

    if '#' in source_url:
        without_fragment = source_url[:source_url.index('#')]
    else:
        without_fragment = source_url

    # git
    if parsed.scheme.startswith('git+') or parsed.path.endswith('.git'):
        # git+anything://anything or anything.git is a git repo:
        name_match = re.match('^.*?([.a-z0-9_-]+?)(.git)?$', parsed.path, re.I)
        if name_match:
            return name_match.group(1), VersionSource('git', without_fragment, parsed.fragment)

    # mercurial
    if parsed.scheme.startswith('hg+') or parsed.path.endswith('.hg'):
        # hg+anything://anything or anything.hg is a hg repo:
        name_match = re.match('^.*?([.a-z0-9_-]+?)(.hg)?$', parsed.path, re.I)
        if name_match:
            return name_match.group(1), VersionSource('hg', without_fragment, parsed.fragment)

    return None, None


def parseSourceURL(source_url):
    ''' Parse the specified version source URL (or version spec), and return an
        instance of VersionSource
    '''
    name, spec = _getNonRegistryRef(source_url)
    if spec:
        return spec

    try:
        url_is_spec = version.Spec(source_url)
    except ValueError:
        url_is_spec = None

    if url_is_spec is not None:
        # if the url is an unadorned version specification (including an empty
        # string) then the source is the module registry:
        return VersionSource('registry', '', source_url)

    raise InvalidVersionSpec("Invalid version specification: \"%s\"" % (source_url))


def isValidSpec(spec_or_source_url):
    ''' Check if the specified version source URL (or version spec), can be
        parsed successfully.
    '''
    try:
        parseSourceURL(spec_or_source_url)
        return True
    except InvalidVersionSpec:
        return False


def parseTargetNameAndSpec(target_name_and_spec):
    ''' Parse targetname[@versionspec] and return a tuple
        (target_name_string, version_spec_string).

        targetname[,versionspec] is also supported (this is how target names
        and specifications are stored internally, and was the documented way of
        setting the spec on the commandline)

        Also accepts raw github version specs (Owner/reponame#whatever), as the
        name can be deduced from these.

        Note that the specification split from the name is not validated. If
        there is no specification (just a target name) passed in, then '*' will
        be returned as the specification.
    '''
    import re
    # fist check if this is a raw github specification that we can get the
    # target name from:
    name, spec = _getNonRegistryRef(target_name_and_spec)
    if name:
        return name, target_name_and_spec

    # next split at the first @ or , if any
    split_at = '@'
    if target_name_and_spec.find('@') > target_name_and_spec.find(',') and \
            ',' in target_name_and_spec:
        split_at = ','
    name = target_name_and_spec.split(split_at)[0]
    spec = target_name_and_spec[len(name)+1:]

    name = name.strip()

    # if there's no specification, return the explicit any-version
    # specification:
    if not spec:
        spec = '*'

    return name, spec

def parseModuleNameAndSpec(module_name_and_spec):
    ''' Parse modulename[@versionspec] and return a tuple
        (module_name_string, version_spec_string).

        Also accepts raw github version specs (Owner/reponame#whatever), as the
        name can be deduced from these.

        Note that the specification split from the name is not validated. If
        there is no specification (just a module name) passed in, then '*' will
        be returned as the specification.
    '''
    import re
    # fist check if this is a raw github specification that we can get the
    # module name from:
    name, spec = _getNonRegistryRef(module_name_and_spec)
    if name:
        return name, module_name_and_spec

    # next split at the first @, if any
    name = module_name_and_spec.split('@')[0]
    spec = module_name_and_spec[len(name)+1:]

    name = name.strip()

    # if there's no specification, return the explicit any-version
    # specification:
    if not spec:
        spec = '*'

    return name, spec



# Copyright 2014 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

# standard library modules, , ,
from __future__ import print_function
import re
import logging
import os

# colorama, BSD 3-Clause license, cross-platform terminal colours, pip install colorama
import colorama

# settings, , load and save settings, internal
from yotta.lib import settings
# Target, , represents an installed target, internal
from yotta.lib import target
# Component, , represents an installed component, internal
from yotta.lib import component
# fsutils, , misc filesystem utils, internal
from yotta.lib import fsutils

# OK this is a pretty terrible validation regex... should find a proper module
# to do this
Target_RE = re.compile('^('+
    '[a-z]+[a-z0-9+-]*('+
        ',[a-zA-Z0-9-]+/[a-zA-Z0-9-]+' +'|'+ '([a-zA-Z0-9_-]*@)?[a-zA-Z0-9_+-]+://.*' + '|' + '[a-z0-9.-]*'+
    ')?'+
')$')


def addOptions(parser):
    parser.add_argument('set_target', default=None, nargs='?',
        # targetname,versionspec_or_url is supported too
        help='set the build target to this (targetname[@versionspec_or_url])'
    )
    parser.add_argument('-g', '--global', dest='save_global',
        default=False, action='store_true',
        help='set globally (in the per-user settings) instead of locally to this directory'
    )
    parser.add_argument('-n', '--no-install', dest='no_install',
        default=False, action='store_true',
        help='do not immediately download the target description'
    )


def displayCurrentTarget(args):
    if not args.plain:
        DIM    = colorama.Style.DIM       #pylint: disable=no-member
        BRIGHT = colorama.Style.BRIGHT    #pylint: disable=no-member
        GREEN  = colorama.Fore.GREEN      #pylint: disable=no-member
        RED    = colorama.Fore.RED        #pylint: disable=no-member
        RESET  = colorama.Style.RESET_ALL #pylint: disable=no-member
    else:
        DIM = BRIGHT = GREEN = RED = RESET = u''

    line = u''

    c = component.Component(os.getcwd())
    if c.isApplication():
        app_path = c.path
    else:
        app_path = None

    derived_target, errors = target.getDerivedTarget(
        args.target, c.targetsPath(), application_dir=app_path, install_missing=False, shrinkwrap=c.getShrinkwrap()
    )
    for error in errors:
        logging.error(error)

    if derived_target is None:
        line = BRIGHT + RED + args.target + u' missing' + RESET
    else:
        for t in derived_target.hierarchy:
            if len(line):
                line += '\n'
            if t:
                line += t.getName() + DIM + u' ' + str(t.getVersion()) + RESET
                if t.installedLinked():
                    line += GREEN + BRIGHT + u' -> ' + RESET + GREEN + fsutils.realpath(t.path)
            else:
                line += BRIGHT + RED + t.getName() + DIM + u' ' + str(t.getVersion()) + BRIGHT + u' missing'
            line += RESET
        base_spec = t.baseTargetSpec()
        if base_spec:
            # if the last target in the hierarchy has a base spec, then the
            # hierarchy is incomplete:
            line += '\n' + BRIGHT + RED + base_spec.name + u' ' + base_spec.versionReq() + u' missing'

    if u'unicode' in str(type(line)):
        # python 2.7
        print(line.encode('utf-8'))
    else:
        print(line)
    return len(errors)


def execCommand(args, following_args):
    if args.set_target is None:
        return displayCurrentTarget(args)
    else:
        from yotta.lib import sourceparse
        from yotta.lib import validate

        name, spec = sourceparse.parseTargetNameAndSpec(args.set_target)
        if not re.match(validate.Target_Name_Regex, name):
            logging.error('Invalid target name: "%s" should use only a-z 0-9 - and +, and start with a letter.' % name)
            return 1
        if not sourceparse.isValidSpec(spec):
            logging.error('Could not parse target version specification: "%s"' % spec)
            return 1

        # separating the target name and spec is still done with a comma
        # internally (for now at least), although @ is the recommended way to
        # set it:
        t = '%s,%s' % (name, spec)

        settings.setProperty('build', 'target', t, not args.save_global)
        settings.setProperty('build', 'targetSetExplicitly', True, not args.save_global)
        if not args.no_install:
            # if we have a module in the current directory, try to make sure
            # this target is installed
            c = component.Component(os.getcwd())
            if c:
                target, errors = c.satisfyTarget(t)
                for err in errors:
                    logging.error(err)
                if len(errors):
                    logging.error('NOTE: use "yotta link-target" to test a locally modified target prior to publishing.')
                    return 1
        return 0
